package com.daisy.devicecommunicationSDK;

import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.Calendar;

import static com.daisy.devicecommunicationSDK.EnumConstants.ByteIndexOf;
import static com.daisy.devicecommunicationSDK.EnumConstants.MIN_TRANS_SEQ;
import static com.daisy.devicecommunicationSDK.EnumConstants.contains;

class DeviceCommunication {
    //Class variables.
    private int transSeq;

    //Constructors.
    DeviceCommunication() {
        transSeq = MIN_TRANS_SEQ;
        Calendar c = Calendar.getInstance();
        {
            int h     = c.get(Calendar.HOUR);
            int sec   = c.get(Calendar.SECOND);
            transSeq += (h + sec);
        }
    }

    /**
     * Add the first 4 bytes in the lpBuf array.
     *
     * @param lpBuf Array that will be modify.
     * @param CmdID The ID of the command that will be send.
     * @param nLen  Size of the array. (include the first 4 bytes and the command length)
     */
    private void MakeTransStart(char[] lpBuf, int CmdID, int nLen) {
        if (lpBuf.length == 0)
            return;

        lpBuf[EnumConstants.BLOCK_STX_NUM] = (char) EnumConstants.TRANS_START;
        lpBuf[EnumConstants.BLOCK_LEN_NUM] = (char) (0x20 + nLen);

        if (transSeq >= 255) {
            transSeq = MIN_TRANS_SEQ;
        } else {
            transSeq++;
        } 
        lpBuf[EnumConstants.BLOCK_SEQ_NUM] = (char) (transSeq);
        lpBuf[EnumConstants.BLOCK_CMD_ID_NUM] = (char) CmdID;
    }

    /**
     * Add the end byte for the command, the check sum, and the end byte of the array.
     *
     * @param lpBuf Array that will be modify.
     * @param nLen  Current position in the array.
     * @return Makes Check sum and returns the lenght of the char array.
     */
    private int MakeTransEnd(char[] lpBuf, int nLen) {
        if (lpBuf.length == 0 || nLen < 0)
            return 0;

        lpBuf[nLen++] += EnumConstants.HOST_DATA_END;
        int BCC = MakeBCC(lpBuf, nLen);
        PutBCC(BCC, lpBuf, nLen);
        nLen += EnumConstants.BLOCK_BCC_SIZE;
        lpBuf[nLen] = EnumConstants.TRANS_END;

        return nLen;
    }

    /**
     * Makes the Check Sum for the array.
     *
     * @param lpStr Array, which is necessary in order to create the check sum
     * @param iLen  Length of the array.
     * @return The check sum of the lpStr array.
     */
    private int MakeBCC(char[] lpStr, int iLen) {
        if (lpStr.length <= 1)
            return 0;

        int iWrk;
        int nCount = 0;
        for (iWrk = 1; iWrk < iLen; iWrk++) {
            nCount += lpStr[iWrk];
        }

        return (nCount);
    }

    /**
     * Puts check sum in the array.
     *
     * @param bcc  The Check sum.
     * @param buff Array, in which the check sum will be added.
     * @param len  Current length of the array.
     */
    private void PutBCC(int bcc, char[] buff, int len) {
        if (buff.length == 0)
            return;

        for (int i = EnumConstants.BLOCK_BCC_SIZE - 1; i >= 0; i--) {
            buff[i + len] += ((bcc & 0x0F) + 0x30);
            bcc = bcc >> 4;
        }
    }

    byte[] GetByteArray(StringBuilder data, int code) {
        char[] trans = new char[1024];
        int wIndex = 0x04;    // start address of data bytes

        byte[] bytes = new byte[300];
        try {
            bytes = data.toString().getBytes("cp1251");
        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
        }

        int lengthBytes = bytes.length;
        for (int i = 0; i < lengthBytes; i++) {
            if (bytes[i] < 0) {
                trans[wIndex] = (char) ((int) bytes[i] + 256);
            } else {
                trans[wIndex] = (char) bytes[i];
            }
            wIndex++;
        }

        MakeTransStart(trans, code, wIndex);
        wIndex = MakeTransEnd(trans, wIndex);

        byte[] result = new byte[wIndex + 1];
        int lengthResult = result.length;
        for (int i = 0; i < lengthResult; i++) {
            result[i] = (byte) trans[i];
        }

        return result;
    }

    /**
     * Reads the response and extract data from it.
     *
     * @param deviceResponse Device response class.
     * @throws UnsupportedEncodingException If cannot encode the data.
     */
    void AddData(DeviceResponse deviceResponse) throws UnsupportedEncodingException
    {
        if ( deviceResponse == null )
        {
            return;
        }
        byte[] receivedData = deviceResponse.GetResponseByte();
        if ( receivedData == null || receivedData.length == 0)
        {
            deviceResponse.setError(EnumConstants.ERROR_NO_RESPONSE);
            return;
        }
        if (receivedData[0] == EnumConstants.RESPONSE_NAK)
        {
            deviceResponse.setError(EnumConstants.ERROR_NAK_RESPONSE);
            return;
        }
        final int ResponsePosStart = ByteIndexOf(receivedData, (byte) EnumConstants.TRANS_START);
        final int ResponsePosEnd = ByteIndexOf(receivedData, (byte) EnumConstants.TRANS_END);
        final int ResponsePrnDataEnd = ByteIndexOf(receivedData, (byte) EnumConstants.PRN_DATA_END);
        final int ResponseDataEnd = ByteIndexOf(receivedData, (byte) EnumConstants.HOST_DATA_END);
        if (
                ResponsePosStart == -1 ||
                ResponsePosEnd == -1 ||
                ResponsePrnDataEnd == -1 ||
                ResponseDataEnd == -1 ||
                ResponsePrnDataEnd < ResponsePosStart ||
                ResponseDataEnd < ResponsePrnDataEnd ||
                ResponsePosEnd < ResponseDataEnd ||
                (ResponsePosEnd - ResponseDataEnd - 1) != 4
            )
        {
            if (contains(receivedData, (byte) EnumConstants.RESPONSE_WAIT))
            {
                deviceResponse.setResponse(EnumConstants.ERROR_SYN_RESPONSE);
                deviceResponse.clearResponseByte();
            }
            else
            {
                deviceResponse.setError(EnumConstants.ERROR_NOT_CORRECT_FORMAT);
                //deviceResponse.setError(EnumConstants.ERROR_NOT_CORRECT_FORMAT + "\n" +"DATA TO HEX: " + EnumConstants.convertToHex(receivedData));
            }
            return;
        }
        byte[] checkSum = Arrays.copyOfRange(receivedData,ResponseDataEnd + 1, ResponsePosEnd);
        int nDataLen = ResponseDataEnd - ResponsePosStart;
        if ((byte) nDataLen != (byte) (receivedData[ResponsePosStart + EnumConstants.BLOCK_SIZE] - 0x20))
        {
            //without + because it can show two times. ^
            //deviceResponse.setError(EnumConstants.ERROR_NOT_CORRECT_FORMAT);
            deviceResponse.setError(EnumConstants.ERROR_NOT_CORRECT_FORMAT + "\n" +"DATA TO HEX: " + EnumConstants.convertToHex(receivedData));
            return;
        }
        byte[] dataToEnd = Arrays.copyOfRange(receivedData, ResponsePosStart + 1, ResponseDataEnd + 1);
        //Check Sum.
        int BCC = 0;
        for (byte aDataToEnd : dataToEnd)
        {
            if (aDataToEnd < 0)
            {
                BCC += aDataToEnd & 0xFF;
            }
            else
            {
                BCC += aDataToEnd;
            }
        }
        char[] buff = String.format("%04X", BCC).toCharArray();
        for (int i = 0; i < 4; i++)
        {
            if (buff[i] > '9')
            {
                buff[i] -= 0x07;
            }
            if (checkSum[i] != buff[i])
            {
                deviceResponse.setError(EnumConstants.ERROR_NOT_CORRECT_CHECKSUM);
                return;
            }
        }
        // Check Status.
        byte[] status = Arrays.copyOfRange(receivedData,ResponsePrnDataEnd + 1, ResponseDataEnd);
        StringBuilder bytesToBits = new StringBuilder();
        for (byte b1 : status)
        {
            bytesToBits.append(String.format("%8s", Integer.toBinaryString(b1 & 0xFF)).replace(' ', '0'));
        }
        String error = CheckForErrors(bytesToBits.toString().toCharArray());
        if(!error.isEmpty())
        {
            deviceResponse.addError(error);
            String byte3 = bytesToBits.toString().substring(25, 32);
            if (byte3.contains("1"))
            {
                int errorByte3Decimal = Integer.parseInt(byte3, 2);
                byte[] errorByte3Mes = GetErrorMessage(errorByte3Decimal);
                byte[] errorByte3 = Arrays.copyOf(errorByte3Mes, errorByte3Mes.length);
                if (errorByte3.length > 0)
                {
                    deviceResponse.setArrayByte3(errorByte3);
                    deviceResponse.setStatus(EnumErrorType.ERROR_BYTE_3);
                }
            }
        }
        String data = new String(Arrays.copyOfRange(receivedData, ResponsePosStart + EnumConstants.BLOCK_DATA_NUM, ResponsePrnDataEnd),
                "Windows-1251");
        deviceResponse.setResponse(data);
    } 

    /**
     * Check array for errors.
     *
     * @param arrayChars The array, which will be checked for errors
     * @return Returns the error message.
     */
    private String CheckForErrors(char[] arrayChars) {
        StringBuilder isError = new StringBuilder();
        for (int i = 0; i < arrayChars.length; i++) {
            if (arrayChars[i] == '1' && i % 8 != 0 && EnumConstants.ErrorBytes.intToErrorBytes(i) != null) {
                if (EnumConstants.BytePositionIsError(i)) {
                    isError.append(EnumConstants.ErrorBytes.intToErrorBytes(i)).append("\n");
                }
            }
        }
        final String byte3 = new String(arrayChars).substring(25, 32); // check the 7 bits from 3-th status byte
        if (byte3.contains("1")) {
            final int errorByte3Decimal = Integer.parseInt(byte3, 2);
            isError.append(EnumConstants.Errors3thByte.intToErrors3thBytes(errorByte3Decimal));
        }
        return isError.toString();
    }

    private byte[] GetErrorMessage(int dec) {
        return GetByteArray(new StringBuilder(dec), EnumConstants.CODE_GET_ERR_DESC);
    }
}
